@@ -5,14 +5,12 @@ use std::hash::{Hash, Hasher};
 use crate::lines::LineNumber;
 use crate::syntax::{ChangeKind, Syntax};
 use rustc_hash::FxHashMap;
+use strsim::normalized_levenshtein;
 use Edge::*;
 
 #[derive(Debug, Clone)]
 struct Vertex<'a> {
     lhs_syntax: Option<&'a Syntax<'a>>,
-    // TODO: rather than forking when we see prev novel (more nodes in
-    // memory, bigger nodes), just always consider nodes on the same
-    // line to be 'closer' (lower cost).
     lhs_prev_novel: Option<LineNumber>,
     rhs_syntax: Option<&'a Syntax<'a>>,
     rhs_prev_novel: Option<LineNumber>,
@@ -75,8 +73,6 @@ impl<'a> Eq for OrdVertex<'a> {}
 enum Edge {
     UnchangedNode(u64),
     UnchangedDelimiter(u64),
-    // TODO: consider a replaced Atom edge if the levenshtein distance
-    // is reasonably close?
     ReplacedComment,
     NovelAtomLHS { contiguous: bool },
     NovelAtomRHS { contiguous: bool },
@@ -95,7 +91,7 @@ impl Edge {
             UnchangedDelimiter(depth_difference) => 100 + min(40, *depth_difference),
 
             // Replacing a comment is better than treating it as novel.
-            ReplacedComment => 1500,
+            ReplacedComment => 150,
 
             // Otherwise, we've added/removed a node.
             NovelAtomLHS { contiguous } | NovelAtomRHS { contiguous } => {
@@ -181,7 +177,7 @@ fn shortest_path(start: Vertex) -> Vec<(Edge, Vertex)> {
     res
 }
 
-const NOVEL_TREE_THRESHOLD: usize = 20;
+const NOVEL_TREE_THRESHOLD: u64 = 20;
 
 fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
     let mut res = vec![];
@@ -250,16 +246,23 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
 
         if let (
             Syntax::Atom {
+                content: lhs_content,
                 is_comment: lhs_is_comment,
                 ..
             },
             Syntax::Atom {
+                content: rhs_content,
                 is_comment: rhs_is_comment,
                 ..
             },
         ) = (lhs_syntax, rhs_syntax)
         {
-            if *lhs_is_comment && *rhs_is_comment {
+            // Both sides are comments and their content is reasonably
+            // similar.
+            if *lhs_is_comment
+                && *rhs_is_comment
+                && normalized_levenshtein(lhs_content, rhs_content) > 0.4
+            {
                 res.push((
                     ReplacedComment,
                     Vertex {
@@ -483,10 +486,7 @@ mod tests {
         let lhs = arena.alloc(Atom {
             info: SyntaxInfo {
                 unique_id: Cell::new(0),
-                pos_content_hash: 0,
-                next: Cell::new(None),
-                change: Cell::new(None),
-                num_ancestors: Cell::new(0),
+                ..SyntaxInfo::new(0)
             },
             position: pos_helper(0),
             content: "foo".into(),
@@ -497,10 +497,7 @@ mod tests {
         let rhs = arena.alloc(Atom {
             info: SyntaxInfo {
                 unique_id: Cell::new(1),
-                pos_content_hash: 1,
-                next: Cell::new(None),
-                change: Cell::new(None),
-                num_ancestors: Cell::new(0),
+                ..SyntaxInfo::new(1)
             },
             position: pos_helper(1),
             content: "foo".into(),
@@ -825,4 +822,65 @@ mod tests {
             ]
         );
     }
+    #[test]
+    fn replace_similar_comment() {
+        let arena = Arena::new();
+
+        let lhs: Vec<&Syntax> = vec![Syntax::new_comment(
+            &arena,
+            pos_helper(1),
+            "the quick brown fox",
+        )];
+        init_info(&lhs);
+
+        let rhs: Vec<&Syntax> = vec![Syntax::new_comment(
+            &arena,
+            pos_helper(1),
+            "the quick brown cat",
+        )];
+        init_info(&rhs);
+
+        let start = Vertex {
+            lhs_syntax: lhs.get(0).map(|n| *n),
+            lhs_prev_novel: None,
+            rhs_syntax: rhs.get(0).map(|n| *n),
+            rhs_prev_novel: None,
+        };
+        let route = shortest_path(start);
+
+        let actions = route.iter().map(|(action, _)| *action).collect_vec();
+        assert_eq!(actions, vec![ReplacedComment]);
+    }
+
+    #[test]
+    fn replace_very_different_comment() {
+        let arena = Arena::new();
+
+        let lhs: Vec<&Syntax> = vec![Syntax::new_comment(
+            &arena,
+            pos_helper(1),
+            "the quick brown fox",
+        )];
+        init_info(&lhs);
+
+        let rhs: Vec<&Syntax> = vec![Syntax::new_comment(&arena, pos_helper(1), "foo bar")];
+        init_info(&rhs);
+
+        let start = Vertex {
+            lhs_syntax: lhs.get(0).map(|n| *n),
+            lhs_prev_novel: None,
+            rhs_syntax: rhs.get(0).map(|n| *n),
+            rhs_prev_novel: None,
+        };
+        let route = shortest_path(start);
+
+        let actions = route.iter().map(|(action, _)| *action).collect_vec();
+        assert_eq!(
+            actions,
+            vec![
+                NovelAtomLHS { contiguous: false },
+                NovelAtomRHS { contiguous: false }
+            ]
+        );
+    }
 }
