@@ -1,10 +1,11 @@
+use crate::intervals::Interval;
 use crate::positions::SingleLineSpan;
 use crate::syntax::{aligned_lines, MatchKind, MatchedPos};
+use lazy_static::lazy_static;
 use regex::Regex;
 use std::cmp::{max, min, Ordering};
 use std::collections::HashMap;
 use std::fmt;
-use std::ops::RangeInclusive;
 
 const SPACER: &str = "  ";
 const MAX_GAP: usize = 1;
@@ -14,26 +15,24 @@ const MIN_WIDTH: usize = 35;
 use pretty_assertions::assert_eq;
 
 #[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
-pub struct LineNumber {
-    pub number: usize,
-}
+pub struct LineNumber(pub usize);
 
 impl fmt::Debug for LineNumber {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.write_fmt(format_args!("LineNumber: {}", self.number))
+        f.write_fmt(format_args!("LineNumber: {}", self.0))
     }
 }
 
 impl From<usize> for LineNumber {
     fn from(number: usize) -> Self {
-        LineNumber { number }
+        LineNumber(number)
     }
 }
 
 #[derive(Debug, PartialEq, Eq)]
 pub struct LineGroup {
-    lhs_lines: Option<RangeInclusive<LineNumber>>,
-    rhs_lines: Option<RangeInclusive<LineNumber>>,
+    lhs_lines: Option<Interval<LineNumber>>,
+    rhs_lines: Option<Interval<LineNumber>>,
 }
 
 impl LineGroup {
@@ -48,14 +47,11 @@ impl LineGroup {
         let mut res = vec![];
         match &self.lhs_lines {
             Some(lhs_lines) => {
-                // TODO: it's not possible to represent an empty range
-                // with RangeInclusive and LineNumber. Define a
-                // separate interval type.
-                if lhs_lines.end().number == 0 {
+                if lhs_lines.is_empty() {
                     return vec![];
                 }
 
-                for line in lhs_lines.start().number..=lhs_lines.end().number {
+                for line in lhs_lines.start.0..=lhs_lines.end.0 {
                     res.push(line.into());
                 }
             }
@@ -71,11 +67,11 @@ impl LineGroup {
                 // TODO: it's not possible to represent an empty range
                 // with RangeInclusive and LineNumber. Define a
                 // separate interval type.
-                if rhs_lines.end().number == 0 {
+                if rhs_lines.end.0 == 0 {
                     return vec![];
                 }
 
-                for line in rhs_lines.start().number..=rhs_lines.end().number {
+                for line in rhs_lines.start.0..=rhs_lines.end.0 {
                     res.push(line.into());
                 }
             }
@@ -85,22 +81,18 @@ impl LineGroup {
     }
 
     pub fn pad(&mut self, amount: usize, max_lhs_line: LineNumber, max_rhs_line: LineNumber) {
-        if let Some(lhs_lines) = self.lhs_lines.take() {
-            let (mut start, mut end) = lhs_lines.into_inner();
+        if let Some(Interval { start, end }) = self.lhs_lines.take() {
+            let start = (max(0, start.0 as isize - amount as isize) as usize).into();
+            let end = min(max_lhs_line.0, end.0 + amount).into();
 
-            start = (max(0, start.number as isize - amount as isize) as usize).into();
-            end = min(max_lhs_line.number, end.number + amount).into();
-
-            self.lhs_lines = Some(start..=end);
+            self.lhs_lines = Some(Interval { start, end });
         }
 
-        if let Some(rhs_lines) = self.rhs_lines.take() {
-            let (mut start, mut end) = rhs_lines.into_inner();
+        if let Some(Interval { start, end }) = self.rhs_lines.take() {
+            let start = (max(0, start.0 as isize - amount as isize) as usize).into();
+            let end = min(max_rhs_line.0, end.0 + amount).into();
 
-            start = (max(0, start.number as isize - amount as isize) as usize).into();
-            end = min(max_rhs_line.number, end.number + amount).into();
-
-            self.rhs_lines = Some(start..=end);
+            self.rhs_lines = Some(Interval { start, end });
         }
     }
 
@@ -108,19 +100,19 @@ impl LineGroup {
     /// line?
     fn next_lg_touches(&self, lg: &LineGroup) -> bool {
         if let (Some(self_lines), Some(lg_lines)) = (&self.lhs_lines, &lg.lhs_lines) {
-            let self_end = self_lines.end();
-            let lg_start = lg_lines.start();
+            let self_end = self_lines.end;
+            let lg_start = lg_lines.start;
 
-            if lg_start.number <= self_end.number + 1 {
+            if lg_start.0 <= self_end.0 + 1 {
                 return true;
             }
         }
 
         if let (Some(self_lines), Some(lg_lines)) = (&self.rhs_lines, &lg.rhs_lines) {
-            let self_end = self_lines.end();
-            let lg_start = lg_lines.start();
+            let self_end = self_lines.end;
+            let lg_start = lg_lines.start;
 
-            if lg_start.number <= self_end.number + 1 {
+            if lg_start.0 <= self_end.0 + 1 {
                 return true;
             }
         }
@@ -132,10 +124,11 @@ impl LineGroup {
     /// `lg`. If either side of `lg` does not overlap with self, fill in
     /// the gap.
     fn next_extend(&mut self, lg: &LineGroup) {
-        match &self.lhs_lines {
-            Some(self_lhs) => {
-                if let Some(lg_lines) = &lg.lhs_lines {
-                    self.lhs_lines = Some(*self_lhs.start()..=*lg_lines.end());
+        // TODO: rename to `lg` to `other`.
+        match &mut self.lhs_lines {
+            Some(Interval { end: self_end, .. }) => {
+                if let Some(Interval { end: lg_end, .. }) = &lg.lhs_lines {
+                    *self_end = *lg_end;
                 }
             }
             None => {
@@ -143,10 +136,10 @@ impl LineGroup {
             }
         }
 
-        match &self.rhs_lines {
-            Some(self_rhs) => {
-                if let Some(lg_lines) = &lg.rhs_lines {
-                    self.rhs_lines = Some(*self_rhs.start()..=*lg_lines.end());
+        match &mut self.rhs_lines {
+            Some(Interval { end: self_end, .. }) => {
+                if let Some(Interval { end: lg_end, .. }) = &lg.rhs_lines {
+                    *self_end = *lg_end;
                 }
             }
             None => {
@@ -165,20 +158,20 @@ impl LineGroup {
         };
 
         if let Some(group_lines) = group_lines {
-            let last_group_line = group_lines.end();
+            let last_group_line = group_lines.end;
 
             let match_lines = &mp.pos;
             assert!(!match_lines.is_empty());
-            let first_match_line = match_lines[0].line.number;
+            let first_match_line = match_lines[0].line.0;
 
-            if first_match_line <= last_group_line.number + max_gap {
+            if first_match_line <= last_group_line.0 + max_gap {
                 return true;
             }
         }
         if let (Some(first_opposite), Some(opposite_group_lines)) =
             (mp.prev_opposite_pos.first(), opposite_group_lines)
         {
-            if first_opposite.line.number <= opposite_group_lines.end().number + max_gap {
+            if first_opposite.line.0 <= opposite_group_lines.end.0 + max_gap {
                 return true;
             }
         }
@@ -189,22 +182,28 @@ impl LineGroup {
     fn add_lhs_pos(&mut self, line_spans: &[SingleLineSpan]) {
         if let (Some(first), Some(last)) = (line_spans.first(), line_spans.last()) {
             if let Some(lhs_lines) = &self.lhs_lines {
-                let start = min(*lhs_lines.start(), first.line);
-                let end = max(*lhs_lines.end(), last.line);
-                self.lhs_lines = Some(start..=end);
+                let start = min(lhs_lines.start, first.line);
+                let end = max(lhs_lines.end, last.line);
+                self.lhs_lines = Some(Interval { start, end });
             } else {
-                self.lhs_lines = Some(first.line..=last.line);
+                self.lhs_lines = Some(Interval {
+                    start: first.line,
+                    end: last.line,
+                });
             }
         }
     }
     fn add_rhs_pos(&mut self, line_spans: &[SingleLineSpan]) {
         if let (Some(first), Some(last)) = (line_spans.first(), line_spans.last()) {
             if let Some(rhs_lines) = &self.rhs_lines {
-                let start = min(*rhs_lines.start(), first.line);
-                let end = max(*rhs_lines.end(), last.line);
-                self.rhs_lines = Some(start..=end);
+                let start = min(rhs_lines.start, first.line);
+                let end = max(rhs_lines.end, last.line);
+                self.rhs_lines = Some(Interval { start, end });
             } else {
-                self.rhs_lines = Some(first.line..=last.line);
+                self.rhs_lines = Some(Interval {
+                    start: first.line,
+                    end: last.line,
+                });
             }
         }
     }
@@ -220,53 +219,18 @@ impl LineGroup {
 
     pub fn max_visible_lhs(&self) -> LineNumber {
         match &self.lhs_lines {
-            Some(lhs_lines) => *lhs_lines.end(),
+            Some(lhs_lines) => lhs_lines.end,
             None => 0.into(),
         }
     }
     pub fn max_visible_rhs(&self) -> LineNumber {
         match &self.rhs_lines {
-            Some(rhs_lines) => *rhs_lines.end(),
+            Some(rhs_lines) => rhs_lines.end,
             None => 0.into(),
         }
     }
 }
 
-fn horizontal_concat(
-    left: &str,
-    right: &str,
-    lhs_content_width: usize,
-    lhs_column_width: usize,
-) -> String {
-    let left_str_lines: Vec<&str> = left.lines().collect();
-    let right_str_lines: Vec<&str> = right.lines().collect();
-
-    let mut i = 0;
-    let mut res = String::new();
-    loop {
-        match (left_str_lines.get(i), right_str_lines.get(i)) {
-            (Some(left_line), Some(right_line)) => {
-                res.push_str(left_line);
-                res.push_str(SPACER);
-                res.push_str(right_line);
-            }
-            (Some(left_line), None) => {
-                res.push_str(left_line);
-            }
-            (None, Some(right_line)) => {
-                res.push_str(&" ".repeat(lhs_column_width + lhs_content_width));
-                res.push_str(SPACER);
-                res.push_str(right_line);
-            }
-            (None, None) => break,
-        }
-        res.push('\n');
-        i += 1;
-    }
-
-    res
-}
-
 /// Compare two MatchedPos to see which starts earlier (on either
 /// side).
 fn compare_matched_pos(lhs: &MatchedPos, rhs: &MatchedPos) -> Ordering {
@@ -421,49 +385,35 @@ fn apply_group(
     lhs_column_width: usize,
     rhs_column_width: usize,
 ) -> String {
-    let mut lhs_result = String::new();
-    let mut rhs_result = String::new();
+    let mut result = String::new();
 
     for (lhs_line_num, rhs_line_num) in
         aligned_lines(&group.lhs_lines(), &group.rhs_lines(), lhs_line_matches)
     {
-        // TODO: we could build up a single string rather than
-        // horizontally concatenating afterwards.
-
         match lhs_line_num {
             Some(lhs_line_num) => {
-                lhs_result.push_str(&format_line_num_padded(
-                    lhs_line_num.number,
-                    lhs_column_width,
-                ));
-                lhs_result.push_str(lhs_lines[lhs_line_num.number]);
+                result.push_str(&format_line_num_padded(lhs_line_num.0, lhs_column_width));
+                result.push_str(lhs_lines[lhs_line_num.0]);
             }
             None => {
-                lhs_result.push_str(&" ".repeat(lhs_column_width));
-                lhs_result.push_str(&" ".repeat(lhs_content_width));
+                result.push_str(&" ".repeat(lhs_column_width));
+                result.push_str(&" ".repeat(lhs_content_width));
             }
         }
-        lhs_result.push('\n');
+        result.push_str(SPACER);
 
         match rhs_line_num {
             Some(rhs_line_num) => {
-                rhs_result.push_str(&format_line_num_padded(
-                    rhs_line_num.number,
-                    rhs_column_width,
-                ));
-                rhs_result.push_str(rhs_lines[rhs_line_num.number]);
+                result.push_str(&format_line_num_padded(rhs_line_num.0, rhs_column_width));
+                result.push_str(rhs_lines[rhs_line_num.0]);
             }
             None => {}
         }
-        rhs_result.push('\n');
+
+        result.push('\n');
     }
 
-    horizontal_concat(
-        &lhs_result,
-        &rhs_result,
-        lhs_content_width,
-        lhs_column_width,
-    )
+    result
 }
 
 /// Display all the lines in `lhs` and `rhs` that are mentioned in
@@ -527,8 +477,10 @@ pub struct NewlinePositions {
 
 impl From<&str> for NewlinePositions {
     fn from(s: &str) -> Self {
-        let newline_re = Regex::new("\n").unwrap();
-        let mut positions: Vec<_> = newline_re.find_iter(s).map(|mat| mat.end()).collect();
+        lazy_static! {
+            static ref NEWLINE_RE: Regex = Regex::new("\n").unwrap();
+        }
+        let mut positions: Vec<_> = NEWLINE_RE.find_iter(s).map(|mat| mat.end()).collect();
         positions.insert(0, 0);
 
         NewlinePositions {
@@ -582,7 +534,7 @@ impl NewlinePositions {
     ) -> Vec<SingleLineSpan> {
         let mut res = vec![];
         for pos in self.from_offsets(region_start, region_end) {
-            if pos.line.number == 0 {
+            if pos.line.0 == 0 {
                 res.push(SingleLineSpan {
                     line: start.line,
                     start_col: start.start_col + pos.start_col,
@@ -590,7 +542,7 @@ impl NewlinePositions {
                 });
             } else {
                 res.push(SingleLineSpan {
-                    line: (start.line.number + pos.line.number).into(),
+                    line: (start.line.0 + pos.line.0).into(),
                     start_col: pos.start_col,
                     end_col: pos.end_col,
                 });
@@ -702,11 +654,11 @@ impl MaxLine for String {
 #[test]
 fn str_max_line() {
     let line: String = "foo\nbar".into();
-    assert_eq!(line.max_line().number, 1);
+    assert_eq!(line.max_line().0, 1);
 }
 
 #[test]
 fn empty_str_max_line() {
     let line: String = "".into();
-    assert_eq!(line.max_line().number, 0);
+    assert_eq!(line.max_line().0, 0);
 }
