@@ -1,10 +1,11 @@
 mod dijkstra;
+mod intervals;
 mod lines;
 mod parse;
 mod positions;
 mod style;
 mod syntax;
-use clap::{App, Arg};
+use clap::{App, AppSettings, Arg};
 use std::ffi::OsStr;
 use std::path::Path;
 use typed_arena::Arena;
@@ -22,7 +23,21 @@ fn term_width() -> Option<usize> {
     term_size::dimensions().map(|(w, _)| w)
 }
 
-const VERSION: &'static str = env!("CARGO_PKG_VERSION");
+/// Do these bytes look like a binary format that we can't do anything
+/// useful with?
+fn is_probably_binary(bytes: &[u8]) -> bool {
+    // If more than 20 of the first 1,000 characters are not valid
+    // UTF-8, we assume it's binary.
+    let num_replaced = String::from_utf8_lossy(&bytes)
+        .to_string()
+        .chars()
+        .take(1000)
+        .filter(|c| *c == std::char::REPLACEMENT_CHARACTER)
+        .count();
+    num_replaced > 20
+}
+
+const VERSION: &str = env!("CARGO_PKG_VERSION");
 
 fn main() {
     let matches = App::new("Difftastic")
@@ -35,24 +50,14 @@ fn main() {
                 .takes_value(true)
                 .help("Override the language parser"),
         )
-        .arg(
-            Arg::with_name("LINES")
-                .long("context")
-                .takes_value(true)
-                .help("Number of lines of context (default 3)"),
-        )
         .arg(
             Arg::with_name("COLUMNS")
                 .long("width")
                 .takes_value(true)
                 .help("Override terminal width"),
         )
-        .arg(
-            Arg::with_name("inline")
-                .long("inline")
-                .help("Prefer single column output"),
-        )
         .arg(Arg::with_name("positional_args").multiple(true))
+        .setting(AppSettings::ArgRequiredElseHelp)
         .get_matches();
 
     let args: Vec<_> = matches.values_of_lossy("positional_args").unwrap();
@@ -67,16 +72,36 @@ fn main() {
         _ => panic!("Expected 2 arguments or 7 arguments"),
     };
 
-    let lhs_src = read_or_die(&lhs_path);
-    let rhs_src = read_or_die(&rhs_path);
-
     let syntax_toml = ConfigDir::read_default_toml();
-
     let lang = match Path::new(&display_path).extension() {
         Some(extension) => find_lang(syntax_toml, &OsStr::to_string_lossy(extension)),
         None => None,
     };
 
+    let lhs_bytes = read_or_die(&lhs_path);
+    let rhs_bytes = read_or_die(&rhs_path);
+    let lhs_binary = is_probably_binary(&lhs_bytes);
+    let rhs_binary = is_probably_binary(&rhs_bytes);
+
+    let lang_name = match &lang {
+        _ if lhs_binary || rhs_binary => "binary".to_string(),
+        Some(lang) => lang.name.clone(),
+        None => "plain text".to_string(),
+    };
+    println!("{}", style::header(&display_path, &lang_name));
+
+    if lhs_binary || rhs_binary {
+        return;
+    }
+
+    // TODO: don't replace tab characters inside string literals.
+    let lhs_src = String::from_utf8_lossy(&lhs_bytes)
+        .to_string()
+        .replace("\t", "    ");
+    let rhs_src = String::from_utf8_lossy(&rhs_bytes)
+        .to_string()
+        .replace("\t", "    ");
+
     let terminal_width = match matches.value_of("COLUMNS") {
         Some(width) => width.parse::<usize>().unwrap(),
         None => term_width().unwrap_or(80),
@@ -101,12 +126,6 @@ fn main() {
 
     let lhs_matched_lines = matching_lines(&lhs);
 
-    let lang_name = match &lang {
-        Some(lang) => lang.name.clone(),
-        None => "plain text".to_string(),
-    };
-    println!("{}", style::header(&display_path, &lang_name));
-
     let mut groups = visible_groups(&lhs_positions, &rhs_positions);
     if groups.is_empty() {
         if lang.is_some() {
@@ -122,8 +141,8 @@ fn main() {
     }
     groups = join_overlapping(groups);
 
-    let lhs_column_width = format_line_num(groups.last().unwrap().max_visible_lhs().number).len();
-    let rhs_column_width = format_line_num(groups.last().unwrap().max_visible_rhs().number).len();
+    let lhs_column_width = format_line_num(groups.last().unwrap().max_visible_lhs().0).len();
+    let rhs_column_width = format_line_num(groups.last().unwrap().max_visible_rhs().0).len();
 
     let lhs_formatted_length = lhs_printable_width(&lhs_src, lhs_column_width, terminal_width);
     let rhs_formatted_length = rhs_printable_width(
