@@ -7,9 +7,9 @@ use std::fs;
 use toml::Value;
 use typed_arena::Arena;
 
-pub fn read_or_die(path: &str) -> String {
+pub fn read_or_die(path: &str) -> Vec<u8> {
     match fs::read(path) {
-        Ok(src) => String::from_utf8_lossy(&src).to_string(),
+        Ok(src) => src,
         Err(e) => {
             match e.kind() {
                 std::io::ErrorKind::NotFound => {
@@ -43,7 +43,6 @@ pub struct Language {
     pub name: String,
     extensions: Vec<String>,
     atom_patterns: Vec<Regex>,
-    string_patterns: Vec<Regex>,
     comment_patterns: Vec<Regex>,
     open_delimiter_pattern: Regex,
     close_delimiter_pattern: Regex,
@@ -89,7 +88,6 @@ fn lang_from_value(name: &str, v: &Value) -> Language {
         name: name.into(),
         extensions: as_string_vec(v.get("extensions").unwrap()),
         atom_patterns: as_regex_vec(v.get("atom_patterns").unwrap()),
-        string_patterns: as_regex_vec(v.get("string_patterns").unwrap()),
         comment_patterns: as_regex_vec(v.get("comment_patterns").unwrap()),
         open_delimiter_pattern: as_regex(
             table
@@ -175,18 +173,6 @@ fn parse_from<'a>(
             }
         }
 
-        // TODO: fix duplication with previous loop
-        for pattern in &lang.string_patterns {
-            if let Some(m) = pattern.find(&s[state.str_i..]) {
-                match current_match {
-                    Some((_, prev_m)) if prev_m.start() <= m.start() => {}
-                    _ => {
-                        current_match = Some((LexKind::Atom, m));
-                    }
-                }
-            }
-        }
-
         if let Some(m) = lang.open_delimiter_pattern.find(&s[state.str_i..]) {
             match current_match {
                 Some((_, prev_m)) if prev_m.start() <= m.start() => {}
@@ -516,15 +502,18 @@ mod tests {
             &parse(&arena, "\"\n\"", &lang()),
             &[Syntax::new_atom(
                 &arena,
-                vec![SingleLineSpan {
-                    line: 0.into(),
-                    start_col: 0,
-                    end_col: 1,
-                }, SingleLineSpan {
-                    line: 1.into(),
-                    start_col: 0,
-                    end_col: 1,
-                }],
+                vec![
+                    SingleLineSpan {
+                        line: 0.into(),
+                        start_col: 0,
+                        end_col: 1,
+                    },
+                    SingleLineSpan {
+                        line: 1.into(),
+                        start_col: 0,
+                        end_col: 1,
+                    },
+                ],
                 "\"\n\"",
             )],
         );
@@ -537,23 +526,26 @@ mod tests {
         assert_syntaxes(
             // "\\" "a"
             &parse(&arena, "\"\\\\\" \"a\"", &lang()),
-            &[Syntax::new_atom(
-                &arena,
-                vec![SingleLineSpan {
-                    line: 0.into(),
-                    start_col: 0,
-                    end_col: 4,
-                }],
-                "\"\\\\\"",
-            ), Syntax::new_atom(
-                &arena,
-                vec![SingleLineSpan {
-                    line: 0.into(),
-                    start_col: 5,
-                    end_col: 8,
-                }],
-                "\"a\"",
-            )],
+            &[
+                Syntax::new_atom(
+                    &arena,
+                    vec![SingleLineSpan {
+                        line: 0.into(),
+                        start_col: 0,
+                        end_col: 4,
+                    }],
+                    "\"\\\\\"",
+                ),
+                Syntax::new_atom(
+                    &arena,
+                    vec![SingleLineSpan {
+                        line: 0.into(),
+                        start_col: 5,
+                        end_col: 8,
+                    }],
+                    "\"a\"",
+                ),
+            ],
         );
     }
 
@@ -694,6 +686,34 @@ mod tests {
         );
     }
 
+    #[test]
+    fn test_parse_indented_multiline_comment() {
+        let arena = Arena::new();
+
+        assert_syntaxes(
+            &parse(&arena, "  /* foo\n  bar */", &lang()),
+            // Deliberately construct an Atom directly, because
+            // Syntax::new_comment has the multiline logic.
+            &[&Atom {
+                info: crate::syntax::SyntaxInfo::new(0),
+                content: "/* foo\nbar */".into(),
+                is_comment: true,
+                position: vec![
+                    SingleLineSpan {
+                        line: 0.into(),
+                        start_col: 2,
+                        end_col: 8,
+                    },
+                    SingleLineSpan {
+                        line: 1.into(),
+                        start_col: 2,
+                        end_col: 8,
+                    },
+                ],
+            }],
+        );
+    }
+
     #[test]
     fn test_parse_list() {
         let arena = Arena::new();
