@@ -91,12 +91,7 @@ pub fn apply_colors(s: &str, is_lhs: bool, positions: &[MatchedPos]) -> String {
                 background: None,
                 bold: false,
             },
-            MatchKind::Moved => Style {
-                foreground: if is_lhs { Color::Red } else { Color::Green },
-                background: None,
-                bold: false,
-            },
-            MatchKind::Novel => Style {
+            MatchKind::Novel | MatchKind::ChangedCommentPart => Style {
                 foreground: if is_lhs {
                     Color::BrightRed
                 } else {
@@ -105,6 +100,11 @@ pub fn apply_colors(s: &str, is_lhs: bool, positions: &[MatchedPos]) -> String {
                 background: None,
                 bold: true,
             },
+            MatchKind::UnchangedCommentPart => Style {
+                foreground: if is_lhs { Color::Red } else { Color::Green },
+                background: None,
+                bold: false,
+            },
         };
         for line_pos in &pos.pos {
             styles.push((*line_pos, style));
