@@ -1,9 +1,10 @@
 #![allow(clippy::mutable_key_type)] // Hash for Syntax doesn't use mutable fields.
 
 use itertools::{EitherOrBoth, Itertools};
+use lazy_static::lazy_static;
 use regex::Regex;
 use std::cell::Cell;
-use std::cmp::min;
+use std::cmp::{max, min};
 use std::collections::hash_map::DefaultHasher;
 use std::collections::HashMap;
 use std::fmt;
@@ -38,6 +39,7 @@ impl<'a> fmt::Debug for ChangeKind<'a> {
 
 /// Fields that are common to both `Syntax::List` and `Syntax::Atom`.
 pub struct SyntaxInfo<'a> {
+    // TODO: Make these fields private.
     pub pos_content_hash: u64,
     pub next: Cell<Option<&'a Syntax<'a>>>,
     pub change: Cell<Option<ChangeKind<'a>>>,
@@ -45,6 +47,18 @@ pub struct SyntaxInfo<'a> {
     pub unique_id: Cell<u64>,
 }
 
+impl<'a> SyntaxInfo<'a> {
+    pub fn new(pos_content_hash: u64) -> Self {
+        Self {
+            pos_content_hash,
+            next: Cell::new(None),
+            change: Cell::new(None),
+            num_ancestors: Cell::new(0),
+            unique_id: Cell::new(0),
+        }
+    }
+}
+
 pub enum Syntax<'a> {
     List {
         info: SyntaxInfo<'a>,
@@ -53,9 +67,7 @@ pub enum Syntax<'a> {
         children: Vec<&'a Syntax<'a>>,
         close_position: Vec<SingleLineSpan>,
         close_content: String,
-        // TODO: this probably makes more sense as a u64, since we
-        // don't use it for indexing.
-        num_descendants: usize,
+        num_descendants: u64,
     },
     Atom {
         info: SyntaxInfo<'a>,
@@ -66,15 +78,13 @@ pub enum Syntax<'a> {
 }
 
 fn dbg_pos(pos: &[SingleLineSpan]) -> String {
-    if pos.is_empty() {
-        "-".into()
-    } else {
-        let start = pos.first().unwrap();
-        let end = pos.last().unwrap();
-        format!(
+    match pos {
+        [] => "-".into(),
+        [pos] => format!("{}:{}-{}", pos.line.0, pos.start_col, pos.end_col),
+        [start, .., end] => format!(
             "{}:{}-{}:{}",
-            start.line.number, start.start_col, end.line.number, end.end_col
-        )
+            start.line.0, start.start_col, end.line.0, end.end_col
+        ),
     }
 }
 
@@ -132,6 +142,21 @@ impl<'a> fmt::Debug for Syntax<'a> {
     }
 }
 
+fn trim_left(max_trim: usize, content: &str, pos: SingleLineSpan) -> (String, SingleLineSpan) {
+    let chars: Vec<_> = content.chars().collect();
+
+    match chars.iter().position(|c| *c != ' ' && *c != '\t') {
+        Some(first_non_whitespace) => {
+            let skip_num = max(max_trim, first_non_whitespace);
+
+            let mut new_pos = pos;
+            new_pos.start_col += skip_num;
+            (chars.iter().skip(skip_num).collect(), new_pos)
+        }
+        None => (content.to_string(), pos),
+    }
+}
+
 impl<'a> Syntax<'a> {
     #[allow(clippy::mut_from_ref)] // Clippy doesn't understand arenas.
     pub fn new_list(
@@ -163,13 +188,7 @@ impl<'a> Syntax<'a> {
         }
 
         arena.alloc(List {
-            info: SyntaxInfo {
-                unique_id: Cell::new(0),
-                pos_content_hash: hasher.finish(),
-                next: Cell::new(None),
-                change: Cell::new(None),
-                num_ancestors: Cell::new(0),
-            },
+            info: SyntaxInfo::new(hasher.finish()),
             open_position,
             open_content: open_content.into(),
             close_content: close_content.into(),
@@ -194,7 +213,27 @@ impl<'a> Syntax<'a> {
         position: Vec<SingleLineSpan>,
         content: &str,
     ) -> &'a mut Syntax<'a> {
-        Self::new_atom_(arena, position, content, true)
+        // Ignore leading whitespace in multiline comments, so changes
+        // in comment indentation are ignored.
+        let first_line_indent = match position.first() {
+            Some(line_pos) => line_pos.start_col,
+            None => 0,
+        };
+
+        let mut new_lines: Vec<String> = vec![];
+        let mut new_position = vec![];
+        for (i, (line, span)) in content.lines().zip(position).enumerate() {
+            if i == 0 {
+                new_lines.push(line.to_string());
+                new_position.push(span);
+            } else {
+                let (new_line, new_span) = trim_left(first_line_indent, line, span);
+                new_lines.push(new_line);
+                new_position.push(new_span);
+            }
+        }
+
+        Self::new_atom_(arena, new_position, &new_lines.join("\n"), true)
     }
 
     #[allow(clippy::mut_from_ref)] // Clippy doesn't understand arenas.
@@ -210,13 +249,7 @@ impl<'a> Syntax<'a> {
         content.hash(&mut hasher);
 
         arena.alloc(Atom {
-            info: SyntaxInfo {
-                unique_id: Cell::new(0),
-                pos_content_hash: hasher.finish(),
-                next: Cell::new(None),
-                change: Cell::new(None),
-                num_ancestors: Cell::new(0),
-            },
+            info: SyntaxInfo::new(hasher.finish()),
             position,
             content: content.into(),
             is_comment,
@@ -456,15 +489,13 @@ fn split_comment_words(
 ) -> Vec<MatchedPos> {
     // TODO: also split on whitespace, so "// (foo)" splits before "(".
 
-    // TODO: don't highlight individual word changes if the lines are
-    // largely novel (e.g. compute levenshtein distance).
-
     // TODO: merge adjacent single-line comments unless there are
     // blank lines between them.
-    let word_boundary = Regex::new(r"\b").unwrap();
-
-    let content_parts: Vec<_> = word_boundary.split(content).collect();
-    let other_parts: Vec<_> = word_boundary.split(opposite_content).collect();
+    lazy_static! {
+        static ref WORD_BOUNDARY_RE: Regex = Regex::new(r"\b").unwrap();
+    }
+    let content_parts: Vec<_> = WORD_BOUNDARY_RE.split(content).collect();
+    let other_parts: Vec<_> = WORD_BOUNDARY_RE.split(opposite_content).collect();
 
     let content_newlines = NewlinePositions::from(content);
 
@@ -714,17 +745,51 @@ pub fn aligned_lines(
     rhs_lines: &[LineNumber],
     lhs_line_matches: &HashMap<LineNumber, LineNumber>,
 ) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
-    let mut rhs_highest_matched = rhs_lines.first().map(|l| l.number as isize).unwrap_or(0) - 1;
+    // When adding padding to a LineGroup where each side has a
+    // different number of lines, we can end up with extra padding on
+    // the side with fewer lines.
+    //
+    // Work around that by discarding any lines at the beginning with
+    // aren't matched.
+    //
+    // TODO: fix padding to be smarter.
+    //
+    // TODO: do the same for the end of the hunk.
+    aligned_lines_(lhs_lines, rhs_lines, lhs_line_matches)
+        .iter()
+        .skip_while(|(lhs, rhs)| lhs.is_none() || rhs.is_none())
+        .copied()
+        .collect()
+}
+
+/// Given two slices of contiguous line numbers, return pairs of
+/// matched lines.
+///
+/// A LHS line is matched with a RHS line if it's present in
+/// `lhs_line_matches` and hasn't already been matched.
+///
+/// If a line has no match on the other side, the pair will contain
+/// None on the other side.
+fn aligned_lines_(
+    lhs_lines: &[LineNumber],
+    rhs_lines: &[LineNumber],
+    lhs_line_matches: &HashMap<LineNumber, LineNumber>,
+) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
+    let mut rhs_highest_matched = rhs_lines.first().map(|l| l.0 as isize).unwrap_or(0) - 1;
 
     // For every LHS line, if there is a RHS line that is included in
     // `rhs_lines` and hasn't yet been paired up, add it to
     // matched_lines.
+    //
+    // TODO: prefer the lie with the most similarity, not just the
+    // first line.  See spurious `let` alignment in 9c71298f8294ce8f,
+    // LHS line 96 in lines.rs.
     let mut matched_lines = vec![];
     for lhs_line in lhs_lines {
         if let Some(rhs_line) = lhs_line_matches.get(lhs_line) {
-            if rhs_line.number as isize > rhs_highest_matched {
+            if rhs_line.0 as isize > rhs_highest_matched {
                 matched_lines.push((lhs_line, rhs_line));
-                rhs_highest_matched = rhs_line.number as isize;
+                rhs_highest_matched = rhs_line.0 as isize;
             }
         }
     }
@@ -829,7 +894,7 @@ mod tests {
         line_matches.insert(2.into(), 12.into());
 
         assert_eq!(
-            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
+            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
             vec![
                 (Some(1.into()), None),
                 (Some(2.into()), Some(12.into())),
@@ -838,6 +903,20 @@ mod tests {
         );
     }
 
+    #[test]
+    fn test_aligned_middle_drop_none_at_start() {
+        let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
+        let rhs_lines: Vec<LineNumber> = vec![12.into(), 13.into()];
+
+        let mut line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
+        line_matches.insert(2.into(), 12.into());
+
+        assert_eq!(
+            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
+            vec![(Some(2.into()), Some(12.into())), (None, Some(13.into()))]
+        );
+    }
+
     #[test]
     fn test_aligned_all() {
         let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
@@ -848,7 +927,7 @@ mod tests {
         line_matches.insert(2.into(), 12.into());
 
         assert_eq!(
-            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
+            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
             vec![
                 (Some(1.into()), Some(11.into())),
                 (Some(2.into()), Some(12.into())),
@@ -864,7 +943,7 @@ mod tests {
         let line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
 
         assert_eq!(
-            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
+            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
             vec![(Some(1.into()), Some(11.into()))]
         );
     }
@@ -879,7 +958,7 @@ mod tests {
         line_matches.insert(2.into(), 11.into());
 
         assert_eq!(
-            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
+            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
             vec![(Some(1.into()), Some(11.into())), (Some(2.into()), None)]
         );
     }
@@ -894,7 +973,7 @@ mod tests {
         line_matches.insert(1.into(), 12.into());
 
         assert_eq!(
-            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
+            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
             vec![
                 (None, Some(11.into())),
                 (Some(1.into()), Some(12.into())),
@@ -912,7 +991,7 @@ mod tests {
         line_matches.insert(1.into(), 10.into());
 
         assert_eq!(
-            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
+            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
             vec![
                 (Some(1.into()), Some(11.into())),
                 (Some(2.into()), Some(12.into())),
@@ -929,7 +1008,7 @@ mod tests {
         line_matches.insert(0.into(), 0.into());
 
         assert_eq!(
-            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
+            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
             vec![(Some(0.into()), Some(0.into()))]
         );
     }
@@ -983,12 +1062,10 @@ mod tests {
         assert_eq!(
             Atom {
                 info: SyntaxInfo {
-                    unique_id: Cell::new(0),
-                    pos_content_hash: 0,
-                    next: Cell::new(None),
                     change: Cell::new(Some(Novel)),
-                    num_ancestors: Cell::new(0),
+                    ..SyntaxInfo::new(1)
                 },
+
                 position: vec![SingleLineSpan {
                     line: 1.into(),
                     start_col: 2,
@@ -999,11 +1076,8 @@ mod tests {
             },
             Atom {
                 info: SyntaxInfo {
-                    unique_id: Cell::new(1),
-                    pos_content_hash: 1,
-                    next: Cell::new(None),
                     change: Cell::new(None),
-                    num_ancestors: Cell::new(1),
+                    ..SyntaxInfo::new(1)
                 },
                 position: vec![SingleLineSpan {
                     line: 1.into(),
@@ -1015,9 +1089,4 @@ mod tests {
             }
         );
     }
-
-    // #[test]
-    // fn test_split_comment_words() {
-
-    // }
 }
